#include <MsTimer2.h>
#include <Servo.h>

Servo myservo[20];

char cmd[101]; //buffer for readed command
int cmdlen;    //length of readed command

char pinModes[22]; //array to keep selected pin modes

byte errNum; //Error flag returned by getNum function - value: 0 or 1
byte posNum; //Next position after readed number - used by getNum function to return position of next character after readed number

byte drvControl=1; //Set off to free pins D5..D8 from DrvControl for servo & gpio
byte dirLpin=8;
byte dirRpin=7;
byte pwmLpin=6;
byte pwmRpin=5;

int servoPosition[20];
int servoTarget[20];
int servoTimer[20];

int i; //for any function internal loops only, may be overwriten in any other function

void error_cmd(int error_id=0)
{
  Serial.print("ERR: Wrong cmd=[");
  for(i=0; i<cmdlen; i++) Serial.print(cmd[i]);
  Serial.print("] error=");
  Serial.println(error_id,DEC);
};

int getNum(int startPos)
{
  char sign='+';
  int res=0;
  int sp;
  sp=startPos;
  if(cmd[sp]=='-' || cmd[sp]=='+'){ sign=cmd[sp]; sp++; };
  if( cmd[sp]<'0' || cmd[sp]>'9'){ errNum=1; return 0; };
  while( cmd[sp]>='0' && cmd[sp]<='9' && sp<101) res=res*10+(cmd[sp++]-'0');
  posNum=sp;
  errNum=0;
  //Serial.println("FnRdNm");
  if(sign=='-') return -res;
  return res;
};

void exec_servo_cmd()
{
  int st=0;
  int pin=0;
  int servoPos[20];
  int servoTime=0;
  for(i=0;i<20;i++)servoPos[i]=0;
  while(true){
    if(cmd[st]==10){
      Serial.println("SrvDo");
      for(i=0;i<20;i++)if(servoPos[i]!=0){
        if(servoTime==0){
          Serial.print('!');
          myservo[i].writeMicroseconds(servoPos[i]);
          servoPosition[i]=servoPos[i];
          servoTarget[i]=servoPos[i];
        }else{
          servoTarget[i]=servoPos[i];
          servoTimer[i]=servoTime;
        };
      };
      return;
    };
    if(cmd[st]=='#'){
      pin=getNum(st+1);
      if( pin<2 || pin>19){ Serial.println("ERR: ServoPin allowed #2..19"); return; };
      if( drvControl==1 && pin>=5 && pin<=8 ){ Serial.println("ERR: ServoPin #5..8 reserved by DrvControl=1"); return; };
      Serial.print("Servo ");
      Serial.println(pin,DEC);
      st=posNum;
      continue;
    };
    if(cmd[st]=='P'){
      servoPos[pin]=getNum(st+1);
      if(pinModes[pin]!='S'){
        pinModes[pin]=='S';
        pinMode(pin,OUTPUT);
        myservo[pin].attach(pin);
      };
      myservo[pin].writeMicroseconds(servoPos[pin]);
      Serial.print("Pos ");
      Serial.println(servoPos[pin],DEC);
      st=posNum;
      continue;
    };
    if(cmd[st]=='T'){
      servoTime=getNum(st+1);
      st=posNum;
      continue;
    };
    Serial.println("ERR: Wrong srv cmd!");
    return;
  };
};

void print_pin_id(int pin){
  if(pin>=0 && pin<=13){
    Serial.print("D");
    Serial.print(pin,DEC);
  } else if(pin>=14 && pin<=21){
    Serial.print("A");
    Serial.print((pin-14),DEC);
  }else{
    Serial.print("WrongPinId");
  };
};

void exec_pin_cmd()
{
  int st=1;
  char pinCmd;
  int pin;
  
  if(cmd[1]=='i' && cmd[2]=='n'){ st=3; }; //ignore 'in' in Pin
  pinCmd=cmd[st];
  st++;
  if(pinCmd=='M'){
    //PinMode
    if(cmd[st]=='o' && cmd[st+1]=='d' && cmd[st+2]=='e'){ st+=3; }; //ignore 'ode' in PinMode
  }else if(pinCmd=='S' || pinCmd=='G'){
    //PinSet
    if(cmd[st]=='e' && cmd[st+1]=='t'){ st+=2; }; //ignore 'et' in PinSet/PinGet
  }else{
    Serial.println(pinCmd);
    error_cmd(1);
    return;
  };
  if(cmd[st]!='A' && cmd[st]!='D' ){ Serial.println(cmd[st]); error_cmd(2); return; };

  pin=getNum(st+1); if(cmd[st]=='A') pin+=14;
  if( errNum ){ error_cmd(3); return; };
  st=posNum;
  if( pin<0 || pin>21){ Serial.println("ERR: WrongPin, allowed only D2..D13, A0..A7"); return; };
  if( drvControl==1 && pin>=5 && pin<=8 ){ Serial.println("ERR: WrongPin, D5..D8 reserved by DrvControl=1"); return; };
  if(pinCmd=='G'){
    if(pinModes[pin]=='A'){
      i=analogRead(pin-14);
      print_pin_id(pin);
      Serial.print("=");
      Serial.print(i,DEC);
      Serial.println("");
      return;
    };
    i=digitalRead(pin);
    print_pin_id(pin);
    Serial.print("=");
    Serial.print(i,DEC);
    Serial.println("");
    return;
  };
  
  if(cmd[st]!='='){ error_cmd(4); return; }; //for PinMode & PinSet commands must be = after pin id
  st++;
  if(pinCmd=='M'){
    if((cmd[st]=='I' || cmd[st]=='O') && (pin<2 || pin>19)){ Serial.println("ERR: PinMode 'In'|'Out' only for D2..D13, A0..A5"); return; };
    if(cmd[st]=='A' && (pin<14 || pin>21)){ Serial.println("ERR: PinMode 'Adc' only for A0..A7"); return; };
    Serial.print("PinMode");
    print_pin_id(pin);
    Serial.print("=");
    if(pinModes[pin]=='S' && cmd[st]!='S') myservo[pin].detach();
    pinModes[pin]=cmd[st];
    if(cmd[st]=='A'){
      if(pin<20) pinMode(pin,INPUT);
      Serial.println("A");
    }else if(cmd[st]=='I'){
      pinMode(pin,INPUT);
      Serial.println("I");
    }else if(cmd[st]=='O'){
      pinMode(pin,OUTPUT);
      Serial.println("O");
    }else if(cmd[st]=='S'){
      pinMode(pin,OUTPUT);
      myservo[pin].attach(pin);
      Serial.println("S");
    };
    return;
  };
  
  if(pinCmd=='S'){
    if(pinModes[pin]=='A'){ Serial.println("ERR: PinSet not allowed in Adc-mode"); return; };
    if(cmd[st]!='0' && cmd[st]!='1'){ Serial.println("ERR: PinSet value must be 0|1"); return; };
    if(cmd[st]=='0') digitalWrite(pin,LOW);
    if(cmd[st]=='1') digitalWrite(pin,HIGH);
    print_pin_id(pin);
    Serial.print("=");
    Serial.println(cmd[st]);
    return;
  };
};

void exec_drv_cmd()
{
  int spdL,spdR,speedL,speedR;
  int st=1;
  if(cmd[1]=='r' && cmd[2]=='v'){ st=3; }; //ignore 'rv' in Drv
  if(cmd[st]=='L'){
    if(drvControl==0){ Serial.println("ERR: DrvLR doesn't work, DrvControl=0"); return; };
    st++;
    if(cmd[st]=='R'){ st++; }; //ignore 'R' in DrvLR
    if(cmd[st]!='='){ error_cmd(8); return; }; //for DrvLR commands must be = after command
    speedL=getNum(st+1);
    if(errNum){ error_cmd(9); return; };
    if(cmd[posNum]!=','){ error_cmd(10); return; }; //for DrvLR commands must be , after first parameter
    speedR=getNum(posNum+1);
    if(errNum){ error_cmd(10); return; };
    if(speedL<-100 || speedL>100 || speedR<-100 || speedR>100){ Serial.println("ERR: DrvLR param range -100..+100"); return; };
    Serial.print("DrvLR=");
    Serial.print(speedL,DEC);
    Serial.print(',');
    Serial.println(speedR,DEC);
    if(speedL<0){ digitalWrite(8,HIGH); }else digitalWrite(8,LOW);
    if(speedR<0){ digitalWrite(7,HIGH); }else digitalWrite(7,LOW);
    spdL=speedL; if(speedL<0) spdL=-speedL;
    spdR=speedR; if(speedR<0) spdR=-speedR;
    if(spdL==100){ digitalWrite(9,HIGH); }else if(spdL==0){ digitalWrite(9,LOW); }else analogWrite(9,spdL+spdL+(spdL>>1));
    if(spdR==100){ digitalWrite(10,HIGH); }else if(spdR==0){ digitalWrite(10,LOW); }else analogWrite(10,spdR+spdR+(spdR>>1));
  }else{
    error_cmd(6);
  };
};

void execute_cmd()
{
  if(cmdlen==0) return;
  for(i=cmdlen; i<101; i++) cmd[i]=10;
  Serial.print("ExecCmd=[");
  for(i=0; i<cmdlen; i++) Serial.print(cmd[i]);
  Serial.println("]");
  
  if( cmdlen==1 && cmd[0]=='?'){
    Serial.println("Help");
  //  Serial.println("  ? - Show this help");
  //  Serial.println("  # - Servo control");
  //  Serial.println("  PinMode - Set pin mode - In/Out/Adc");
  //  Serial.println("  PinGet - Get pin value (result depends on mode)");
  //  Serial.println("  PinSet - Set pin value (accepted value depends on mode");
  //  Serial.println("  DrvLR - DC Motors control");
    return;
  };
  
  if( cmd[0]=='#' ){ exec_servo_cmd(); Serial.println("SrvDn"); return; };
  if( cmd[0]=='P' ){ exec_pin_cmd(); Serial.println("PinDn"); return; };
  if( cmd[0]=='D' ){ exec_drv_cmd(); Serial.println("DrvDn"); return; };
  
  error_cmd(7);
}

void updateServos(){
  for(i=0; i<20; i++)if(pinModes[i]=='S')if(servoTarget[i]!=servoPosition[i])
  {
    if(servoTimer[i]<=25){ servoPosition[i]=servoTarget[i]; servoTimer[i]=0; }else{ servoPosition[i]=servoPosition[i]+(servoTarget[i]-servoPosition[i])*25/servoTimer[i]; servoTimer[i]-=25; };
    myservo[i].writeMicroseconds(servoPosition[i]);
  };
}

void setup()
{
   Serial.begin(19200);
   cmdlen=0;
   for(i=2; i<=13; i++) pinModes[i]='I';
   for(i=14; i<=21; i++) pinModes[i]='A';
   for(i=0; i<20; i++) servoPosition[i]=1500;
   MsTimer2::set(25, updateServos); // 25ms period
   //MsTimer2::start();
}

void loop()
{
  while (Serial.available() > 0) {
    char data=Serial.read();
    if( data==10 ){ execute_cmd(); cmdlen=0; }
    else if( data<33 ) continue;
    else{ cmd[cmdlen++]=data; };
    if( cmdlen==100 ){
      Serial.println("ERR: Cmd long, 100bytes max.");
      cmdlen=0;
    };
  };
}
